local addonName, DruidBuffs = ...

----------------------------------------------------------------------------------------------------
-- settings / constants / references
----------------------------------------------------------------------------------------------------
DruidBuffsSave = nil -- saved settings - defaults are set up in ADDON_LOADED

local CHECK_SECONDS = 5 -- how often to check for needed buffs

-- key binding things
_G["BINDING_HEADER_DRUIDBUFFS"]                                    = "Druid Buffs"
_G["BINDING_NAME_CLICK DruidBuffsCastBuff:LeftButton"]             = "Cast Next Buff"
_G["BINDING_NAME_CLICK DruidBuffsToggleThornsOnSelf:LeftButton"]   = "Toggle using Thorns on self"
_G["BINDING_NAME_CLICK DruidBuffsToggleThornsOnTarget:LeftButton"] = "Toggle using Thorns on target"

-- spell/icon IDs to check if the player has them
local MARK_OF_THE_WILD_ID     = 1126
local GIFT_OF_THE_WILD_ID1    = 21849
local GIFT_OF_THE_WILD_ID2    = 21850
local GIFT_OF_THE_WILD_ID3    = 26991
local OMEN_OF_CLARITY_ID      = 16864
local THORNS_ID               = 467
local PREPARATION_ID          = 44521
local ARENA_PREPARATION_ID1   = 32727
local ARENA_PREPARATION_ID2   = 32728
local SPIRIT_HEAL_ID          = 44535

-- get locale names for spells
local MARK_OF_THE_WILD = (GetSpellInfo(MARK_OF_THE_WILD_ID))
local GIFT_OF_THE_WILD = (GetSpellInfo(GIFT_OF_THE_WILD_ID1))
local OMEN_OF_CLARITY  = (GetSpellInfo(OMEN_OF_CLARITY_ID))
local THORNS           = (GetSpellInfo(THORNS_ID))
local PHASE_SHIFT      = (GetSpellInfo(4511))

-- local references to commonly used functions
local GetRaidRosterInfo  = GetRaidRosterInfo
local IsInGroup          = IsInGroup
local IsInRaid           = IsInRaid
local IsSpellInRange     = IsSpellInRange
local IsSpellKnown       = IsSpellKnown
local UnitAura           = UnitAura
local UnitInRange        = UnitInRange
local UnitIsConnected    = UnitIsConnected
local UnitIsDeadOrGhost  = UnitIsDeadOrGhost
local UnitName           = UnitName

-- list for temporarily ignoring players with auras that are better so can't be replaced
local tooPowerfulMarks = {} -- tooPowerfulMarks[name] = GetTime() + rebuff time
local tooPowerfulThorns = {}

----------------------------------------------------------------------------------------------------
-- setting the next buff
----------------------------------------------------------------------------------------------------
-- hidden secure buttons used with the key binding - its target/spell is set up in SetNextBuff()
local buttonCastBuff = CreateFrame("Button", "DruidBuffsCastBuff", nil, "SecureActionButtonTemplate")
buttonCastBuff:SetAttribute("type", "spell")

-- counter for periodic buff checking
local checkCountdown = 2 -- make the first check fast

-- helper function to check for the Preparation/Spirit Heal buffs in battlegrounds and arenas
local function HasFreeSpellAura()
	local _, id
	for i=1,100 do
		_, _, _, _, _, _, _, _, _, id = UnitAura("player", i)
		if not id then
			break
		end
		if id == PREPARATION_ID or id == SPIRIT_HEAL_ID or id == ARENA_PREPARATION_ID1 or id == ARENA_PREPARATION_ID2 then
			return true
		end
	end
end

--------------------------------------------------
-- return which buff someone needs
--------------------------------------------------
function DruidBuffs:FindBuffNeeded(name, checkDistance, locationSettings)
	if (checkDistance and IsSpellInRange(MARK_OF_THE_WILD, name) ~= 1) or not UnitIsConnected(name) or UnitIsDeadOrGhost(name) then
		return nil
	end

	local hasOmen, hasMark, hasThorns
	local rebuffOmen, rebuffMark, rebuffThorns

	local buff, expires, _
	local hasFreeAura = HasFreeSpellAura()
	local checkThorns = (locationSettings.checkThorns and DruidBuffsSave.thornsList[name]) or (locationSettings.freeThorns and hasFreeAura)
	local checkMark = locationSettings.checkMark
	local isPlayer = UnitIsUnit("player", name)

	for i=1,100 do
		buff, _, _, _, _, expires = UnitAura(name, i)
		if not buff then
			break
		end

		local left = expires - GetTime()

		if (buff == MARK_OF_THE_WILD or buff == GIFT_OF_THE_WILD) then
			hasMark = true
			if left < locationSettings.rebuffMark or (locationSettings.freeRebuff and hasFreeAura and left < 1620) then
				rebuffMark = true
			end
		elseif buff == THORNS then
			hasThorns = true
			if left < locationSettings.rebuffThorns or (locationSettings.freeRebuff and hasFreeAura and left < 420) then
				rebuffThorns = true
			end
		elseif buff == OMEN_OF_CLARITY then
			hasOmen = true
			if left < locationSettings.rebuffOmen or (locationSettings.freeRebuff and hasFreeAura and left < 1620) then
				rebuffOmen = true
			end
		elseif buff == PHASE_SHIFT then
			return nil -- imp can't receive buffs now
		end
	end

	if isPlayer and (not hasOmen or rebuffOmen) and IsSpellKnown(OMEN_OF_CLARITY_ID) then
		return OMEN_OF_CLARITY
	end

	if checkMark and IsSpellKnown(MARK_OF_THE_WILD_ID) then
		if not hasMark or (rebuffMark and (not tooPowerfulMarks[name] or GetTime() > tooPowerfulMarks[name])) then
			return MARK_OF_THE_WILD
		end
	end

	if checkThorns and IsSpellKnown(THORNS_ID) then
		if not hasThorns or (rebuffThorns and (not tooPowerfulThorns[name] or GetTime() > tooPowerfulThorns[name])) then
			return THORNS
		end
	end
end

--------------------------------------------------
-- return true if target's subgroup needs Gift of the Wild
--------------------------------------------------
function DruidBuffs:GroupNeedsGift(target, targetGroup, locationSettings)
	if not locationSettings.useGift then
		return
	end

	local usable3, noMana3 = IsUsableSpell(GIFT_OF_THE_WILD_ID3)
	local usable2, noMana2 = IsUsableSpell(GIFT_OF_THE_WILD_ID2)
	local usable1, noMana1 = IsUsableSpell(GIFT_OF_THE_WILD_ID1)
	if (not usable3 and not noMana3) and (not usable2 and not noMana2) and (not usable1 and not noMana1) then
		return
	elseif locationSettings.giftAmount == 1 then
		return true
	elseif not targetGroup then
		return
	end

	local countPets = locationSettings.giftCountPets
	local checkRange = locationSettings.giftCheckRange
	local giftAmount = locationSettings.giftAmount

	local count = 1 -- how many need Mark of the Wild in the group
	if IsInRaid() then
		local name, _, subgroup
		for i=1,MAX_RAID_MEMBERS do
			name, _, subgroup = GetRaidRosterInfo(i)
			if name and targetGroup == subgroup then
				if target ~= name and (not checkRange or UnitInRange(name)) and self:FindBuffNeeded(name, false, locationSettings) == MARK_OF_THE_WILD then
					count = count + 1
					if count >= giftAmount then
						return true
					end
				end

				if countPets then
					local petName = name.."-Pet"
					if target ~= petName and (not checkRange or UnitInRange(petName)) and self:FindBuffNeeded(petName, false, locationSettings) == MARK_OF_THE_WILD then
						count = count + 1
						if count >= giftAmount then
							return true
						end
					end
				end
			end
		end
	else
		-- the player won't be in the party member loop, so check now
		if target ~= UnitName("player") and self:FindBuffNeeded(UnitName("player"), false, locationSettings) == MARK_OF_THE_WILD then
			count = count + 1
			if count >= giftAmount then
				return true
			end
		end
		local name
		for i=1,GetNumSubgroupMembers() do
			name = UnitName("party"..i)
			if target ~= name and (not checkRange or UnitInRange(name)) and self:FindBuffNeeded(name, false, locationSettings) == MARK_OF_THE_WILD then
				count = count + 1
				if count >= giftAmount then
					return true
				end
			end

			if countPets then
				local petName = name.."-Pet"
				if target ~= petName and (not checkRange or UnitInRange(petName)) and self:FindBuffNeeded(petName, false, locationSettings) == MARK_OF_THE_WILD then
					count = count + 1
					if count >= giftAmount then
						return true
					end
				end
			end
		end
	end
end

--------------------------------------------------
-- set up the next buff
--------------------------------------------------
function DruidBuffs:SetNextBuff()
	checkCountdown = CHECK_SECONDS
	if InCombatLockdown() then
		return
	end

	local locationSettings = DruidBuffsSave[DruidBuffs:GetLocation()]
	if not locationSettings.enabled then
		DruidBuffs.buffText:SetText("")
		buttonCastBuff:SetAttribute("spell", nil)
		return
	end

	local playerName = UnitName("player")
	local target, spell
	local group
	local needsThornsName, petNeedsMark, petNeedsThorns

	-- check self first
	spell = self:FindBuffNeeded(playerName, true, locationSettings)
	if spell then
		target = playerName
		-- find the player's group number
		if IsInRaid() then
			local name, _
			for i=1,MAX_RAID_MEMBERS do
				name, _, group = GetRaidRosterInfo(i)
				if name == playerName then
					break
				end
			end
		elseif IsInGroup() then
			group = 1
		end
	else
		-- check raid or party members - Mark of the Wild is wanted first, so only stop if that is
		-- found to be needed. If someone needs Thorns, then save their name in case no one else
		-- needed Mark of the Wild. Pets are lowest priority, so save their name too

		-- raid
		if IsInRaid() then
			local name, _
			for i=1,MAX_RAID_MEMBERS do
				name, _, group = GetRaidRosterInfo(i)
				if name and name ~= playerName then
					-- first the raid member
					spell = self:FindBuffNeeded(name, true, locationSettings)
					if spell then
						if spell == MARK_OF_THE_WILD then
							target = name
							break
						elseif not needsThornsName then
							needsThornsName = name
						end
						spell = nil
					-- then their pet
					elseif not petNeedsMark then
						name = name .. "-Pet"
						spell = self:FindBuffNeeded(name, true, locationSettings)
						if spell then
							if spell == MARK_OF_THE_WILD then
								petNeedsMark = name
							elseif not petNeedsThorns then
								petNeedsThorns = name
							end
							spell = nil
						end
					end
				end
			end
		-- party
		elseif IsInGroup() then
			local name
			for i=1,GetNumSubgroupMembers() do
				name = UnitName("party"..i)
				if name ~= playerName then
					-- first the party member
					spell = self:FindBuffNeeded(name, true, locationSettings)
					if spell then
						if spell == MARK_OF_THE_WILD then
							target = name
							group = 1
							break
						elseif not needsThornsName then
							needsThornsName = name
						end
						spell = nil
					-- then their pet
					elseif not petNeedsMark then
						name = name .. "-Pet"
						spell = self:FindBuffNeeded(name, true, locationSettings)
						if spell then
							if spell == MARK_OF_THE_WILD then
								petNeedsMark = name
							elseif not petNeedsThorns then
								petNeedsThorns = name
							end
							spell = nil
						end
					end
				end
			end
		end

		-- last resort spells if nothing more important was found
		if not spell then
			if needsThornsName then
				spell = THORNS
				target = needsThornsName
			elseif petNeedsMark then
				spell = MARK_OF_THE_WILD
				target = petNeedsMark
			elseif petNeedsThorns then
				spell = THORNS
				target = petNeedsThorns
			end
		end
	end

	-- upgrade to Gift of the Wild if needed
	if spell == MARK_OF_THE_WILD and self:GroupNeedsGift(target, group, locationSettings) then
		spell = GIFT_OF_THE_WILD
	end

	-- set the buff button and text - don't show the text while in safe areas
	buttonCastBuff:SetAttribute("spell", spell)
	buttonCastBuff:SetAttribute("unit", target)
	if spell and target and (GetZonePVPInfo()) ~= "sanctuary" and locationSettings.showText then
		if DruidBuffsSave.removeRealmName then
			local realm = target:match("%-[^%-]+")
			if realm and realm ~= "-Pet" then
				target = target:gsub(realm, "")
			end
		end
		DruidBuffs.buffText:SetText(target .. ": " .. spell)
	else
		DruidBuffs.buffText:SetText("")
	end
end

--------------------------------------------------
-- check buffs periodically
--------------------------------------------------
local eventFrame = CreateFrame("frame")
eventFrame:SetScript("OnUpdate", function(self, elapsed)
	checkCountdown = checkCountdown - elapsed
	if checkCountdown <= 0 then
		DruidBuffs:SetNextBuff()
	end
end)

--------------------------------------------------
-- handle events that affect buff checks
--------------------------------------------------
local savedCastTarget, savedCastGuid, savedCastSpell

eventFrame:SetScript("OnEvent", function(self, event, arg1, arg2, arg3, arg4)
	if not InCombatLockdown() then
		--------------------------------------------------
		-- checking for aura changes
		--------------------------------------------------
		if event == "COMBAT_LOG_EVENT_UNFILTERED" then
			local _, action, _, _, _, _, _, _, _, _, _, _, spellId = CombatLogGetCurrentEventInfo()
			if action == "SPELL_CAST_SUCCESS" or action == "SPELL_AURA_APPLIED" or action == "SPELL_AURA_REFRESH" then
				local spellname = (GetSpellInfo(spellId))
				if spellname == MARK_OF_THE_WILD or spellName == THORNS or spellName == OMEN_OF_CLARITY then
					checkCountdown = 0
				end
			end
			return
		end

		if event == "UNIT_AURA" then
			if UnitIsUnit(arg1, "player") or UnitInParty(arg1) or UnitInRaid(arg1) or arg1:find("^partypet") or arg1:find("^raidpet") then
				if checkCountdown > 0.05 then
					checkCountdown = 0.05
				end
			end
			return
		end

		if event == "UNIT_SPELLCAST_SUCCEEDED" then
			local spellName = (GetSpellInfo(arg3))
			if spellName == MARK_OF_THE_WILD or spellName == THORNS or spellName == OMEN_OF_CLARITY then
				checkCountdown = 0
			end
		end

		--------------------------------------------------
		-- stronger aura tracking - save mark/thorn spells
		--------------------------------------------------
		if event == "UNIT_SPELLCAST_SENT" then
			if arg1 == "player" then
				local spellName = (GetSpellInfo(arg4))
				if spellName == MARK_OF_THE_WILD or spellName == THORNS then
					if UnitIsOtherPlayersPet(arg2) then
						savedCastTarget = DruidBuffs:ConvertPetToUnit(arg2)
					else
						savedCastTarget = arg2
					end
					savedCastGuid = arg3
					savedCastSpell = spellName
				end
			end
			return
		end

		-- part of stronger aura tracking - cancel it for any other error
		if event == "UI_ERROR_MESSAGE" then
			if arg2 ~= SPELL_FAILED_AURA_BOUNCED then
				savedCastGuid = nil
			end
			return
		end

		-- part of stronger aura tracking - should come after the UI error
		if event == "UNIT_SPELLCAST_FAILED" then
			if savedCastGuid and savedCastGuid == arg2 then
				local locationSettings = DruidBuffsSave[DruidBuffs:GetLocation()]
				if savedCastSpell == MARK_OF_THE_WILD then
					tooPowerfulMarks[savedCastTarget] = GetTime() + locationSettings.rebuffMark
				else
					tooPowerfulThorns[savedCastTarget] = GetTime() + locationSettings.rebuffThorns
				end
				savedCastGuid = nil
				DruidBuffs:SetNextBuff()
			end
			return
		end
	end

	-- combat started, so disable buff checking and hide the text
	if event == "PLAYER_REGEN_DISABLED" then
		eventFrame:UnregisterEvent("UNIT_AURA")
		DruidBuffs.buffText:Hide()
		eventFrame:Hide()
		return
	end

	-- combat ended, so restart buff checking and reshow the text
	if event == "PLAYER_REGEN_ENABLED" then
		eventFrame:RegisterEvent("UNIT_AURA")
		DruidBuffs:SetNextBuff()
		DruidBuffs.buffText:Show()
		eventFrame:Show()
		return
	end

	-- when leaving a group, remove the too powerful buff data - if still in a group, check buffs fast
	if event == "GROUP_ROSTER_UPDATE" then
		if IsInGroup() then
			if checkCountdown > 0.1 then
				checkCountdown = 0.1
			end
		else
			tooPowerfulMarks = {}
			tooPowerfulThorns = {}
		end
		return
	end

	-- check right after some events
	if event == "PLAYER_ALIVE" then
		if checkCountdown > 0.1 then
			checkCountdown = 0.1
		end
		return
	end

	-- the addon has loaded, so set default settings if needed
	if event == "ADDON_LOADED" and arg1 == addonName then
		eventFrame:UnregisterEvent(event)

		if DruidBuffsSave                 == nil then DruidBuffsSave                 = {} end
		if DruidBuffsSave.textSize        == nil then DruidBuffsSave.textSize        = 16 end
		if DruidBuffsSave.textColor       == nil then DruidBuffsSave.textColor       = {1, 0.49, 0.04} end
		if DruidBuffsSave.thornsList      == nil then DruidBuffsSave.thornsList      = {[UnitName("player")]=true} end
		if DruidBuffsSave.removeRealmName == nil then DruidBuffsSave.removeRealmName = false end

		for _,location in ipairs(DruidBuffs.LocationList) do
			if not DruidBuffsSave[location] then
				DruidBuffsSave[location] = {}
			end
			locationSave = DruidBuffsSave[location]

			-- upgrading - copy old settings
			if locationSave.enabled == nil and DruidBuffsSave.checkThorns ~= nil then
				locationSave.checkThorns  = DruidBuffsSave.checkThorns
				locationSave.rebuffMark   = DruidBuffsSave.rebuffMark
				locationSave.rebuffThorns = DruidBuffsSave.rebuffThorns
				locationSave.rebuffOmen   = DruidBuffsSave.rebuffOmen

				if (location == "World" or location == "Instance") and DruidBuffsSave.giftAmount >= 1 and DruidBuffsSave.giftAmount <= 5 then
					locationSave.useGift = true
					locationSave.giftAmount = DruidBuffsSave.giftAmount
				else
					locationSave.useGift = false
					locationSave.giftAmount = 3
				end
			end

			if locationSave.enabled        == nil then locationSave.enabled        = true end
			if locationSave.checkMark      == nil then locationSave.checkMark      = true end
			if locationSave.checkThorns    == nil then locationSave.checkThorns    = true end
			if locationSave.showText       == nil then locationSave.showText       = true end
			if locationSave.useGift        == nil then locationSave.useGift        = (location == "Instance") end
			if locationSave.giftAmount     == nil then locationSave.giftAmount     = 3    end
			if locationSave.giftCountPets  == nil then locationSave.giftCountPets  = true end
			if locationSave.giftCheckRange == nil then locationSave.giftCheckRange = (location ~= "Instance") end
			if location == "Battleground" or location == "Arena" then
				if locationSave.rebuffMark   == nil then locationSave.rebuffMark   = 0 end
				if locationSave.rebuffThorns == nil then locationSave.rebuffThorns = 0 end
				if locationSave.rebuffOmen   == nil then locationSave.rebuffOmen   = 0 end
			else
				if locationSave.rebuffMark   == nil then locationSave.rebuffMark   = 4*60 end
				if locationSave.rebuffThorns == nil then locationSave.rebuffThorns = 2*60 end
				if locationSave.rebuffOmen   == nil then locationSave.rebuffOmen   = 4*60 end
			end
			if locationSave.freeThorns == nil then locationSave.freeThorns = (location == "Battleground" or location == "Arena") end
			if locationSave.freeRebuff == nil then locationSave.freeRebuff = (location == "Battleground" or location == "Arena") end
		end

		-- upgrading - remove old settings
		if DruidBuffsSave.checkThorns ~= nil then
			DruidBuffsSave.checkThorns  = nil
			DruidBuffsSave.giftAmount   = nil
			DruidBuffsSave.rebuffMark   = nil
			DruidBuffsSave.rebuffThorns = nil
			DruidBuffsSave.rebuffOmen   = nil
		end

		DruidBuffs:LoadBuffText()
		return
	end
end)
eventFrame:RegisterEvent("ADDON_LOADED")                -- temporary - to set up default settings
eventFrame:RegisterEvent("PLAYER_ALIVE")                -- to check for buffs fast after coming back to life
eventFrame:RegisterEvent("PLAYER_REGEN_DISABLED")       -- to know when combat starts
eventFrame:RegisterEvent("PLAYER_REGEN_ENABLED")        -- to know when combat ends
eventFrame:RegisterEvent("GROUP_ROSTER_UPDATE")         -- to clear saved buff times when leaving a group or check for buffs faster
-- to know when a buff/debuff has changed - unit_aura and combat_log_event_unfiltered aren't 100%
-- reliable and seem to miss events in certain situations - not sure about unit_spellcast_succeeded
eventFrame:RegisterEvent("UNIT_AURA")
eventFrame:RegisterEvent("COMBAT_LOG_EVENT_UNFILTERED")
eventFrame:RegisterEvent("UNIT_SPELLCAST_SUCCEEDED")
-- for handling more powerful aura errors
eventFrame:RegisterEvent("UNIT_SPELLCAST_SENT")         -- saves target and guid
eventFrame:RegisterEvent("UI_ERROR_MESSAGE")            -- sees the actual reason
eventFrame:RegisterEvent("UNIT_SPELLCAST_FAILED")       -- takes action if guid still matches
