local _, DruidBuffs = ...

----------------------------------------------------------------------------------------------------
-- Helper functions
----------------------------------------------------------------------------------------------------
-- return a checkbox widget
local function CreateCheckbox(owner, name, text, tooltip)
	local frame = CreateFrame("CheckButton", name, owner, "OptionsCheckButtonTemplate")
	_G[frame:GetName().."Text"]:SetText(text)
	local width = _G[frame:GetName().."Text"]:GetStringWidth()
	frame:SetHitRectInsets(0, -width, 4, 4)
	frame.tooltipText = tooltip
	return frame
end

local function ValidatedNumber(value, min, max, default)
	local number = tonumber(value)
	if not number then
		return default
	elseif number < min then
		return min
	elseif number > max then
		return max
	end
	return number
end

----------------------------------------------------------------------------------------------------
-- Main settings
----------------------------------------------------------------------------------------------------
DruidBuffs.settingsPanelMain = CreateFrame("frame", "DruidBuffsSettingsMain", UIParent)
DruidBuffs.settingsPanelMain.name = "Druid Buffs"
InterfaceOptions_AddCategory(DruidBuffs.settingsPanelMain)

local mainPanel = DruidBuffs.settingsPanelMain

do
	local panel = mainPanel

	local textTitle = mainPanel:CreateFontString(nil, "ARTWORK", "GameFontNormalLarge")
	textTitle:SetPoint("TOPLEFT", panel, "TOPLEFT", 16, -16)
	textTitle:SetText("Druid Buffs")

	local textDescription = panel:CreateFontString(nil, "ARTWORK", "GameFontWhiteSmall")
	textDescription:SetPoint("TOPLEFT", textTitle, "BOTTOMLEFT", 0, -8)
	textDescription:SetText("Each location (world, instances, battlegrounds, and arenas) have their own settings.")

	--------------------------------------------------
	-- Font section
	--------------------------------------------------
	local textBuffExample -- forward declaration

	local textFontHeader = panel:CreateFontString(nil, "ARTWORK", "GameFontNormal")
	textFontHeader:SetPoint("TOPLEFT", textDescription, "BOTTOMLEFT", 0, -24)
	textFontHeader:SetText("Buff text settings:")

	-- Remove realm names
	local cbRemoveRealmName = CreateCheckbox(panel, "DruidBuffsCbRemoveRealmName", "|cffffffffRemove realm names|r")
	cbRemoveRealmName:SetPoint("TOPLEFT", textFontHeader, "BOTTOMLEFT", -2, -8)
	cbRemoveRealmName:SetScript("OnClick", function(self)
		DruidBuffsSave.removeRealmName = self:GetChecked() or false
		DruidBuffs:SetNextBuff()
	end)

	-- Text size
	local textFontSize = panel:CreateFontString(nil, "ARTWORK", "GameFontNormal")
	textFontSize:SetPoint("TOPLEFT", cbRemoveRealmName, "BOTTOMLEFT", 2, -8)
	textFontSize:SetText("|cffffffffSize:|r")

	local editboxFontSize = CreateFrame("EditBox", "DruidBuffsEditBoxFontSize", panel, "InputBoxTemplate")
	editboxFontSize:SetSize(32, 12)
	editboxFontSize:SetPoint("LEFT", textFontSize, "RIGHT", 16, 0)
	editboxFontSize:SetNumeric(true)
	editboxFontSize:SetMaxLetters(2)
	editboxFontSize:SetAutoFocus(false)
	editboxFontSize:SetScript("OnEnterPressed", function(self) self:ClearFocus() end)
	editboxFontSize:SetScript("OnEditFocusLost", function(self)
		local size = ValidatedNumber(self:GetText(), 6, 72, 16)
		DruidBuffsSave.textSize = size
		self:SetText(size)
		DruidBuffs:UpdateTextSize()
		textBuffExample:SetFont("Fonts/FRIZQT__.TTF", DruidBuffsSave.textSize, "OUTLINE")
	end)

	-- Text color
	local buttonFontColor = CreateFrame("Button", "DruidBuffsButtonFontColor", panel, "UIPanelButtonTemplate")
	buttonFontColor:SetSize(90, 20)
	buttonFontColor:SetPoint("TOPLEFT", textFontSize, "BOTTOMLEFT", 0, -12)
	_G[buttonFontColor:GetName().."Text"]:SetText("Set Color")
	buttonFontColor:SetScript("OnClick", function() DruidBuffs:ShowColorPicker() end)

	-- Example text (already declared local)
	textBuffExample = panel:CreateFontString(nil, "ARTWORK", "GameFontNormal")
	textBuffExample:SetText(UnitName("player") .. ": Mark of the Wild")
	textBuffExample:SetPoint("LEFT", buttonFontColor, "RIGHT", 8, 0)
	DruidBuffs.exampleText = textBuffExample

	-- Loading the panel
	panel:SetScript("OnShow", function(self)
		editboxFontSize:SetText(DruidBuffsSave.textSize)
		textBuffExample:SetTextColor(unpack(DruidBuffsSave.textColor))
		textBuffExample:SetFont("Fonts/FRIZQT__.TTF", DruidBuffsSave.textSize, "OUTLINE")
		cbRemoveRealmName:SetChecked(DruidBuffsSave.removeRealmName)
	end)

	panel:Hide()
end

----------------------------------------------------------------------------------------------------
-- Location settings
----------------------------------------------------------------------------------------------------
for _,location in ipairs(DruidBuffs.LocationList) do
	local panelName = "settingsPanel" .. location
	DruidBuffs[panelName] = CreateFrame("frame", "DruidBuffs" .. panelName, DruidBuffs.settingsPanelMain)
	local panel = DruidBuffs[panelName]
	panel.name = location
	panel.parent = DruidBuffs.settingsPanelMain.name
	InterfaceOptions_AddCategory(panel)

	local panelTitle = panel:CreateFontString(nil, "ARTWORK", "GameFontNormalLarge")
	panelTitle:SetPoint("TOPLEFT", panel, "TOPLEFT", 16, -16)
	panelTitle:SetText(location)

	--------------------------------------------------
	-- Main Section
	--------------------------------------------------
	local cbEnabled = CreateCheckbox(panel, "DruidBuffsCbEnabled" .. location, "|cffffffffEnable buff checking in this location|r")
	cbEnabled:SetPoint("TOPLEFT", panelTitle, "BOTTOMLEFT", -2, -8)
	cbEnabled:SetScript("OnClick", function(self)
		DruidBuffsSave[location].enabled = self:GetChecked() or false
		DruidBuffs:SetNextBuff()
	end)

	local cbCheckThorns = CreateCheckbox(panel, "DruidBuffsCbCheckThorns" .. location, "|cffffffffCheck for Thorns|r")
	cbCheckThorns:SetPoint("TOPLEFT", cbEnabled, "BOTTOMLEFT", 0, -8)
	cbCheckThorns:SetScript("OnClick", function(self)
		DruidBuffsSave[location].checkThorns = self:GetChecked() or false
		DruidBuffs:SetNextBuff()
	end)

	local cbCheckMark = CreateCheckbox(panel, "DruidBuffsCbCheckMark" .. location, "|cffffffffCheck for Mark of the Wild|r")
	cbCheckMark:SetPoint("TOPLEFT", cbCheckThorns, "BOTTOMLEFT", 0, -8)
	cbCheckMark:SetScript("OnClick", function(self)
		DruidBuffsSave[location].checkMark = self:GetChecked() or false
		DruidBuffs:SetNextBuff()
	end)

	local cbShowText = CreateCheckbox(panel, "DruidBuffsCbShowText" .. location, "|cffffffffShow the needed buff text|r", "You can still use the key binding to cast the next spell.")
	cbShowText:SetPoint("TOPLEFT", cbCheckMark, "BOTTOMLEFT", 0, -8)
	cbShowText:SetScript("OnClick", function(self)
		DruidBuffsSave[location].showText = self:GetChecked() or false
		DruidBuffs:SetNextBuff()
	end)

	--------------------------------------------------
	-- Gift of the Wild section
	--------------------------------------------------
	local editboxGiftAmount -- forward declaration

	local textGiftHeader = panel:CreateFontString(nil, "ARTWORK", "GameFontNormal")
	textGiftHeader:SetPoint("TOPLEFT", cbShowText, "BOTTOMLEFT", 2, -24)
	textGiftHeader:SetText("Gift of the Wild settings:")

	local cbUseGift = CreateCheckbox(panel, "DruidBuffsCbUseGift" .. location, "|cffffffffUse Gift of the Wild if this many need MotW:|r")
	cbUseGift:SetPoint("TOPLEFT", textGiftHeader, "BOTTOMLEFT", -2, -8)
	cbUseGift:SetScript("OnClick", function(self)
		DruidBuffsSave[location].useGift = self:GetChecked() or false
		DruidBuffs:SetNextBuff()
	end)

	-- already declared local
	editboxGiftAmount = CreateFrame("EditBox", "DruidBuffsEditBoxGiftAmount", panel, "InputBoxTemplate")
	editboxGiftAmount:SetSize(32, 12)
	editboxGiftAmount:SetPoint("LEFT", _G[cbUseGift:GetName().."Text"], "RIGHT", 8, 0)
	editboxGiftAmount:SetNumeric(true)
	editboxGiftAmount:SetMaxLetters(2)
	editboxGiftAmount:SetAutoFocus(false)
	editboxGiftAmount:SetScript("OnEnterPressed", function(self) self:ClearFocus() end)
	editboxGiftAmount:SetScript("OnEditFocusLost", function(self)
		local amount = ValidatedNumber(self:GetText(), 1, 10, 3)
		self:SetText(amount)
		DruidBuffsSave[location].giftAmount = amount
		DruidBuffs:SetNextBuff()
	end)

	local cbCountPets = CreateCheckbox(panel, "DruidBuffsCbCountPets" .. location, "|cffffffffWhen counting how many need MotW, include player pets|r")
	cbCountPets:SetPoint("TOPLEFT", cbUseGift, "BOTTOMLEFT", 0, -8)
	cbCountPets:SetScript("OnClick", function(self)
		DruidBuffsSave[location].giftCountPets = self:GetChecked() or false
		DruidBuffs:SetNextBuff()
	end)

	local cbCheckRange = CreateCheckbox(panel, "DruidBuffsCbCheckRange" .. location, "|cffffffffWhen counting how many need MotW, only count those in range|r",
		"This should probably be disabled in instances because people slightly out of range can still get the buff and everyone is usually nearby or will be soon. In battlegrounds, people are often scattered and too far.")
	cbCheckRange:SetPoint("TOPLEFT", cbCountPets, "BOTTOMLEFT", 0, -8)
	cbCheckRange:SetScript("OnClick", function(self)
		DruidBuffsSave[location].giftCheckRange = self:GetChecked() or false
		DruidBuffs:SetNextBuff()
	end)

	--------------------------------------------------
	-- Rebuff section
	--------------------------------------------------
	local textRebuffHeader = panel:CreateFontString(nil, "ARTWORK", "GameFontNormal")
	textRebuffHeader:SetPoint("TOPLEFT", cbCheckRange, "BOTTOMLEFT", 2, -24)
	textRebuffHeader:SetText("Rebuff when this many minutes are left:")

	-- Mark of the Wild
	local textRebuffMark = panel:CreateFontString(nil, "ARTWORK", "GameFontNormal")
	textRebuffMark:SetPoint("TOPLEFT", textRebuffHeader, "BOTTOMLEFT", 0, -8)
	textRebuffMark:SetText("|cffffffffMark of the Wild:|r")

	local editboxRebuffMark = CreateFrame("EditBox", "DruidBuffsEditBoxRebuffMark" .. location, panel, "InputBoxTemplate")
	editboxRebuffMark:SetSize(32, 12)
	editboxRebuffMark:SetPoint("LEFT", textRebuffMark, "LEFT", 130, 0)
	editboxRebuffMark:SetNumeric(true)
	editboxRebuffMark:SetMaxLetters(2)
	editboxRebuffMark:SetAutoFocus(false)
	editboxRebuffMark:SetScript("OnEnterPressed", function(self) self:ClearFocus() end)
	editboxRebuffMark:SetScript("OnEditFocusLost", function(self)
		local minutes = ValidatedNumber(self:GetText(), 0, 59, 4)
		self:SetText(minutes)
		DruidBuffsSave[location].rebuffMark = minutes * 60
		DruidBuffs:SetNextBuff()
	end)

	-- Thorns
	local textRebuffThorns = panel:CreateFontString(nil, "ARTWORK", "GameFontNormal")
	textRebuffThorns:SetPoint("TOPLEFT", textRebuffMark, "BOTTOMLEFT", 0, -8)
	textRebuffThorns:SetText("|cffffffffThorns:|r")

	local editboxRebuffThorns = CreateFrame("EditBox", "DruidBuffsEditBoxRebuffThorns" .. location, panel, "InputBoxTemplate")
	editboxRebuffThorns:SetSize(32, 12)
	editboxRebuffThorns:SetPoint("LEFT", textRebuffThorns, "LEFT", 130, 0)
	editboxRebuffThorns:SetNumeric(true)
	editboxRebuffThorns:SetMaxLetters(2)
	editboxRebuffThorns:SetAutoFocus(false)
	editboxRebuffThorns:SetScript("OnEnterPressed", function(self) self:ClearFocus() end)
	editboxRebuffThorns:SetScript("OnEditFocusLost", function(self)
		local minutes = ValidatedNumber(self:GetText(), 0, 9, 2)
		self:SetText(minutes)
		DruidBuffsSave[location].rebuffThorns = minutes * 60
		DruidBuffs:SetNextBuff()
	end)

	-- Omen of Clarity
	local textRebuffOmen = panel:CreateFontString(nil, "ARTWORK", "GameFontNormal")
	textRebuffOmen:SetPoint("TOPLEFT", textRebuffThorns, "BOTTOMLEFT", 0, -8)
	textRebuffOmen:SetText("|cffffffffOmen:|r")

	local editboxRebuffOmen = CreateFrame("EditBox", "DruidBuffsEditBoxRebuffOmen" .. location, panel, "InputBoxTemplate")
	editboxRebuffOmen:SetSize(32, 12)
	editboxRebuffOmen:SetPoint("LEFT", textRebuffOmen, "LEFT", 130, 0)
	editboxRebuffOmen:SetNumeric(true)
	editboxRebuffOmen:SetMaxLetters(2)
	editboxRebuffOmen:SetAutoFocus(false)
	editboxRebuffOmen:SetScript("OnEnterPressed", function(self) self:ClearFocus() end)
	editboxRebuffOmen:SetScript("OnEditFocusLost", function(self)
		local minutes = ValidatedNumber(self:GetText(), 0, 29, 4)
		self:SetText(minutes)
		DruidBuffsSave[location].rebuffOmen = minutes * 60
		DruidBuffs:SetNextBuff()
	end)

	-- tabbing through editboxes
	editboxRebuffMark:SetScript("OnTabPressed",   function(self) self:HighlightText(0,0) if IsShiftKeyDown() then editboxRebuffOmen:SetFocus()   else editboxRebuffThorns:SetFocus() end end)
	editboxRebuffThorns:SetScript("OnTabPressed", function(self) self:HighlightText(0,0) if IsShiftKeyDown() then editboxRebuffMark:SetFocus()   else editboxRebuffOmen:SetFocus()   end end)
	editboxRebuffOmen:SetScript("OnTabPressed",   function(self) self:HighlightText(0,0) if IsShiftKeyDown() then editboxRebuffThorns:SetFocus() else editboxRebuffMark:SetFocus()   end end)

	--------------------------------------------------
	-- Special location options
	--------------------------------------------------
	local cbFreeThorns
	local cbFreeRebuff
	if location == "Battleground" or location == "Arena" then
		local textSpecialHeader = panel:CreateFontString(nil, "ARTWORK", "GameFontNormal")
		textSpecialHeader:SetPoint("TOPLEFT", textRebuffOmen, "BOTTOMLEFT", 0, -24)
		textSpecialHeader:SetText(location .. " special options:")

		cbFreeThorns = CreateCheckbox(panel, "DruidBuffsCbFreeThorns" .. location, "|cffffffffCast Thorns on everyone during Preparation/Spirit Heal|r",
			"This will check for Thorns while you have those free casting buffs even if the main Thorns checking option is disabled.")
		cbFreeThorns:SetPoint("TOPLEFT", textSpecialHeader, "BOTTOMLEFT", 2, -8)
		cbFreeThorns:SetScript("OnClick", function(self)
			DruidBuffsSave[location].freeThorns = self:GetChecked() or false
			DruidBuffs:SetNextBuff()
		end)

		cbFreeRebuff = CreateCheckbox(panel, "DruidBuffsCbFreeRebuff" .. location, "|cffffffffRebuff much sooner during Preparation|r",
			"This will rebuff Omen of Clarity and MotW with less than 27 minutes left and Thorns with less than 7 minutes left.")
		cbFreeRebuff:SetPoint("TOPLEFT", cbFreeThorns, "BOTTOMLEFT", 0, -8)
		cbFreeRebuff:SetScript("OnClick", function(self)
			DruidBuffsSave[location].freeRebuff = self:GetChecked() or false
			DruidBuffs:SetNextBuff()
		end)
	end

	--------------------------------------------------
	-- Loading the panel
	--------------------------------------------------
	function panel:UpdateSettingDisplay()
		local settings = DruidBuffsSave[location]

		cbEnabled:SetChecked(settings.enabled)
		cbCheckThorns:SetChecked(settings.checkThorns)
		cbCheckMark:SetChecked(settings.checkMark)
		cbShowText:SetChecked(settings.showText)

		cbUseGift:SetChecked(settings.useGift)
		editboxGiftAmount:SetText(settings.giftAmount)
		cbCountPets:SetChecked(settings.giftCountPets)
		cbCheckRange:SetChecked(settings.giftCheckRange)

		editboxRebuffMark:SetText(settings.rebuffMark / 60)
		editboxRebuffThorns:SetText(settings.rebuffThorns / 60)
		editboxRebuffOmen:SetText(settings.rebuffOmen / 60)

		if location == "Battleground" or location == "Arena" then
			cbFreeThorns:SetChecked(settings.freeThorns)
			cbFreeRebuff:SetChecked(settings.freeRebuff)
		end
	end

	panel:SetScript("OnShow", function(self) self:UpdateSettingDisplay() end)

	panel:Hide()
end
